////////////////////////////////////////
///  The Ox code to compute the simlulation results of the MA2SLS estimators
///  reported in "Constructing Optimal Instruments by First Stage Prediction
///  Averaging" by Kuersteiner and Okui.
////////////////////////////////////////


#include <oxstd.h>
#import <maximize>
#import <solveqp>

////////////////////////////////////
///////////	class TSLS	////
////////////////////////////////////
class Tsls
{
	decl vy, vx, mZ, n;
	decl betatsls;

	Tsls(const y, const x, const Z);
	virtual Estimate();
	virtual	Coverage();
	GetEstimate();
}	

Tsls::Tsls(const y, const x, const Z)
{
	vy = y; vx = x; mZ = Z; n = rows(vx);   
}

Tsls::Estimate()
{
	betatsls =invertsym(vx'mZ*invertsym(mZ'mZ)*mZ'vx)*vx'mZ*invertsym(mZ'mZ)*mZ'vy; 
}

Tsls::Coverage()
{
  decl xhat;
  xhat = mZ*invertsym(mZ'mZ)*mZ'vx;
  decl vhat; 
  vhat = (1/n)*(vy- vx*betatsls)' (vy- vx*betatsls)
	 * invertsym(xhat'vx) *(xhat'xhat) *invertsym(vx'xhat);  
 return (0.1 > ( betatsls - 1.96*sqrt(vhat)) ) 
	    && (0.1 < (betatsls + 1.96*sqrt(vhat)));
}

Tsls::GetEstimate()
{
	return betatsls;
}


//////////// Model averaging //////////////////
class MATsls : Tsls
{

	decl weight, k, mK, mU;
	decl sigmau, sigmae, sigmaue ; 
	decl mGmin, mGmax, mxpx, mxpy, mypy;
	MATsls(const y, const x, const Z);
	Firststage();
	Estimate();
	Coverage();
	GetKWp();
	GetKWn();
	
	SelWDN();
	SelWMA();
	SelWUMAN();
	SelWMAN();
	SelWMA01();
	SelWKW();	
}

MATsls::GetKWp()
{
	return (maxc((weight~zeros(k,1))'))*mK;
}
MATsls::GetKWn()
{
	return fabs((minc((weight~zeros(k,1))'))*mK);
}

MATsls::MATsls(const y, const x, const Z)
{
	Tsls(y,x,Z); k = columns(mZ); mK = range(1,k)';
	mxpx = zeros(k,1); mxpy = zeros(k,1); mypy = zeros(k,1);
    mGmin = mK*ones(1,k) .< ones(k,1)*(mK') .? mK*ones(1,k) .: ones(k,1)*(mK');
	mGmax = mK*ones(1,k) .> ones(k,1)*(mK') .? mK*ones(1,k) .: ones(k,1)*(mK');
 	decl j; for (j=0;j<k;j++)
	{
		mxpx[j] = vx'mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vx;
		mxpy[j] = vx'mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vy;
		mypy[j] = vy'mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vy;
	}
	mU = mK*ones(1,k) .> ones(k,1)*(mK') .? mxpx*ones(1,k) .: ones(k,1)*(mxpx');
	mU = vx'vx - mU;
	mU = mU - mU[k-1][k-1]*ones(k,k);
	Firststage();

}

MATsls::Firststage()
{
// Mallow's criterion for the first stage //
   sigmau = (vx'vx - vx'mZ*invertsym(mZ'mZ)*mZ'vx)/n;

     decl dnc; dnc=zeros(k,2); dnc[][0] = range(0,k-1)';
     decl j; for (j=0;j<k;j++){
     dnc[j][1]=mU[j][j] + 2*(j+1)*sigmau;
     }
     dnc = sortbyc(dnc,1); decl optk = dnc[0][0];
     decl vxhat; vxhat =  mZ[][:optk]*invertsym(mZ[][:optk]'mZ[][:optk])*(mZ[][:optk]'vx);
     decl delta;  delta =invertsym(vxhat'vx)*(vxhat'vy);

// minimizing MSE //   
   sigmae = (vy - delta*vx)'(vy - delta*vx)/n; 
   sigmau = (vx - vxhat)'(vx - vxhat)/n;
   sigmaue=(vx - vxhat)'(vy - delta*vx)/n; 

}



MATsls::Estimate()
{
	betatsls = (weight'mxpy) / (weight'mxpx);
}


MATsls::Coverage()
{
	decl xhat=0;
 	decl j; for (j=0;j<k;j++)
	{
		xhat = xhat +weight[j]*mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*(mZ[][:j]'vx);
	}
	decl vhat;
	vhat = (1/n)*(vy- vx*betatsls)' (vy- vx*betatsls)
		 * invertgen(xhat'vx) *(xhat'xhat) *invertgen(vx'xhat);  

	return (0.1 > ( betatsls - 1.96*sqrt(vhat)) ) 
	    && (0.1 < (betatsls + 1.96*sqrt(vhat)));
}

////

MATsls::SelWDN()
{
 decl dnc; dnc=zeros(k,2); dnc[][0] = range(0,k-1)';
 decl j;
   for (j=0; j<k ; j++) {
       dnc[j][1] = (sigmaue^2)*((j+1)^2) + sigmae*(mU[j][j] + sigmau*(j+1));
       }
   dnc = sortbyc(dnc,1); 
   decl optk = dnc[0][0];
   weight = zeros(k,1); weight[optk]=1;
}
////

MATsls::SelWMA()
{


    decl vlo = zeros(k,1); decl vhi = ones(k,1);
    decl ir;
    decl mK = range(1,k)';
    decl mG, vG, mC, vD;
    mG = 2*((sigmaue^2)*mK*(mK') + sigmae*mU - sigmae*sigmau*mGmin);
    vG = 2*sigmau*sigmae*mK;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
 
   [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,vlo,vhi);

}
MATsls::SelWMA01()
{


    decl vlo = zeros(k,1);//-ones(k,1);
	decl vhi = ones(k,1);
    decl ir;
    decl mK = range(1,k)';
    decl mG, vG, mC, vD;
    mG = 2*((sigmaue^2)*mK*(mK') + (sigmaue^2)*mGmin + sigmae*mU  );
    vG = -mK*2*(sigmae*sigmau +4*sigmaue^2 ) + 2*sigmau*sigmae*mK ;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
 
   [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,vlo,vhi);

}


MATsls::SelWMAN()
{


    decl vlo = -ones(k,1);
	decl vhi = ones(k,1);
    decl ir;
    decl mK = range(1,k)';
    decl mG, vG, mC, vD;
    mG = 2*((sigmaue^2)*mK*(mK') + (sigmaue^2)*mGmin + sigmae*mU);
    vG = -mK*2*(sigmae*sigmau + 4*sigmaue^2) + 2*sigmau*sigmae*mK ;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
 
   [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,vlo,vhi);

}
MATsls::SelWUMAN()
{
    decl ir;
    decl mK = range(1,k)';
    decl mG, vG, mC, vD;
    mG = 2*((sigmaue^2)*mK*(mK') + (sigmaue^2)*mGmin + sigmae*mU );
    vG = -mK*2*(sigmae*sigmau + 4*sigmaue^2) + 2*sigmau*sigmae*mK ;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
 
   [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,<>,<>);
}


/// Kernel Wegihted 2SLS ///
MATsls::SelWKW()
{
	 decl dnc; dnc=zeros(k,2); dnc[][0] = range(0,k-1)';

    weight = zeros(k,1);	// starting value
	decl j;
	dnc[][0] = range(0,k-1)';
 	 for (j=0;j<k;j++){
	 weight = zeros(k,1);
	 weight[:j] = 1/(j+1);
 	 dnc[j][1]= sigmaue*sigmaue*((weight'mK)^2)
	 	+ sigmae*(weight'mU*weight + sigmau*2*mK'weight - sigmau*weight'mGmin*weight );
     }

	 dnc = sortbyc(dnc,1); decl optk = dnc[0][0];
   weight = zeros(k,1);	weight[:optk] = 1 / (optk+1);

}




////////////// simulation class ///////

class Monte
{
	decl n, k, c, Rf2,model;
	decl monte, mSigma, beta, gamma;
	decl betaest, covest, kwest, kwnest;
	
	Monte(const sn, const sk, const sc, const r, const smodel);
	Experiment();
	GetResult();
}

Monte::Monte(const sn, const sk, const sc, const r, const smodel)
{
	n = sn; k=sk; c = sc; Rf2 = r; model=smodel;
	monte =5000;
	mSigma = ones(2,2); mSigma[0][1] = c; mSigma [1][0] =c;
	beta =0.1;

	if 	(model == 1) {
	// part 1 //
	 gamma = ones(k,1)*sqrt(Rf2/(k*(1-Rf2 ) ) ) ;
	}
	else if (model ==2)	{
	//part 2 //
	  gamma=zeros(k,1); decl j; 
	  for (j=0;j<k;j++){ gamma[j]=(1- (j/(k+1)))^4; }
	gamma = sqrt(Rf2/((1-Rf2)*(gamma'gamma)) )*gamma;  
	}
	else if (model == 3){
	// part 3 //
	gamma = zeros(k,1);	decl j; 
	  for (j=0;j<k/2;j++){ gamma[j+k/2]=(1- (j/(k/2+1)))^4; }
	gamma = sqrt(Rf2/((1-Rf2)*(gamma'gamma)) )*gamma; 
	}
}

Monte::Experiment()
{
	ranseed(23587608);

// parameter values //
       decl vy, vx, mZ, mU; 
       decl mSigmac; mSigmac = choleski(mSigma); 

print("gammas", gamma'); 
print ("n=" , n, " K= ", k , " c=",mSigma[1][0] , " Rf2 = ", Rf2  );

// monte carlo //

   decl betatsls = zeros(monte,1);
   decl betadntsls = zeros(monte,1);
   decl betama = zeros(monte,1);
   decl betama0 = zeros(monte,1);
   decl betaman = zeros(monte,1);
   decl betauman = zeros(monte,1);
   decl betakw = zeros(monte,1);

	decl kwkw = zeros(monte,1);
	decl kwdn = zeros(monte,1);
	decl kwman = zeros(monte,1);
	decl kwuman = zeros(monte,1);
	decl kwma = zeros(monte,1);
	decl kwma0 = zeros(monte,1);

	decl kwmann = zeros(monte,1);
	decl kwumann = zeros(monte,1);
	
decl i; for(i=0;i<monte;i++){

// DGP //
       mZ = rann(n,k); mU = rann(n,2)*(mSigmac');
       vx = mZ*gamma + mU[][1];
       vy = beta*vx  + mU[][0]; 

// TSLS ESTIMATION //
       decl mybeta; 
       mybeta = new Tsls(vy,vx,mZ);
       mybeta.Estimate();
       betatsls[i] = mybeta.GetEstimate();
       delete mybeta;

// DNTSLS ESTIMATION //
       mybeta = new MATsls(vy,vx,mZ);
	   mybeta.SelWDN();
	   mybeta.Estimate();
       betadntsls[i] = mybeta.GetEstimate();
		kwdn[i] = mybeta.GetKWp();

// KWTSLS ESTIMATION //
	   mybeta.SelWKW();	   
       mybeta.Estimate();
       betakw[i] = mybeta.GetEstimate();
		kwkw[i] = mybeta.GetKWp();

// MA2SLS-U ESTIMATION //
	   mybeta.SelWUMAN();	   
       mybeta.Estimate();
       betauman[i] = mybeta.GetEstimate();
		kwuman[i] =  mybeta.GetKWp();
		kwumann[i] =  mybeta.GetKWn();

// MA2SLS-C ESTIMATION //
	   mybeta.SelWMAN();	   
       mybeta.Estimate();
       betaman[i] = mybeta.GetEstimate();
		kwman[i] =  mybeta.GetKWp();
		kwmann[i] =  mybeta.GetKWn();
// MA2SLS-P ESTIMATION //
	   mybeta.SelWMA();	   
       mybeta.Estimate();
       betama[i] = mybeta.GetEstimate();
		kwma[i] =  mybeta.GetKWp();
 
// MA2SLS-Ps ESTIMATION //
	   mybeta.SelWMA01();	   
       mybeta.Estimate();
       betama0[i] = mybeta.GetEstimate();
		kwma0[i] =  mybeta.GetKWp();
       delete mybeta;

}

betaest = betatsls~betadntsls~betakw~betauman~betaman~betama0~betama;
kwest = kwdn~kwkw~kwuman~kwman~kwma0~kwma;
kwnest = kwumann~kwmann;
}

Monte::GetResult()
{

	decl mad =  quantilec( fabs((betaest) -beta));
	decl iqr = quantilec(betaest,0.75) - quantilec(betaest,0.25);
	decl maddn = mad/mad[][1];
	decl ressta = (quantilec(betaest)-beta)|iqr|mad|maddn|(k~meanc(kwest))|(zeros(1,3)~meanc(kwnest)~zeros(1,2));

	print( "%r", {"median bias","IQR","MAD","RMAD","KW+","KW-"},
	       "%c",
	{"2SLS", "DN","KW","MA-U","MA-C","MA-P","MA-Ps"},
		"%8.3g" ,ressta);
}

//////////// MAIN PROGRAM ///////////
main()
{
	decl na,ma,ca,r2a,moda; decl myexp;
	na=<100>; ma = <20>;ca=<0.1;0.5;0.9>; r2a = <0.1;0.01>; moda=<1;2;3>;
	decl ni,mi,ci,r2i,modi;
		
		for (ni=0; ni < rows(na);ni++){
		for (modi=0; modi < rows(moda);modi++){
		for (r2i=0; r2i < rows(r2a);r2i++){
		for (mi=0; mi < rows(ma);mi++){
		for (ci=0; ci < rows(ca);ci++){
	myexp = new Monte(na[ni],ma[mi],ca[ci],r2a[r2i],moda[modi]);
	myexp.Experiment();
	myexp.GetResult();
	delete myexp;
	}
	}
	}
	}
	}


	na=<1000>; ma = <30>;ca=<0.1;0.5;0.9>; r2a = <0.1;0.01>; moda=<1;2;3>;
		
		for (ni=0; ni < rows(na);ni++){
		for (modi=0; modi < rows(moda);modi++){
		for (r2i=0; r2i < rows(r2a);r2i++){
		for (mi=0; mi < rows(ma);mi++){
		for (ci=0; ci < rows(ca);ci++){
	myexp = new Monte(na[ni],ma[mi],ca[ci],r2a[r2i],moda[modi]);
	myexp.Experiment();
	myexp.GetResult();
	delete myexp;
	}
	}
	}
	}
	}
	
}