////////////////////////////////////////
///  The Ox code to compute the simlulation results of the MALIML estimators
///  reported in "Constructing Optimal Instruments by First Stage Prediction
///  Averaging" by Kuersteiner and Okui.
///  Note: "RMAD" in the output is MAD relative to LIML-DN while "RMAD in
///  the paper is MAD relative to 2SLS-DN.
////////////////////////////////////////


#include <oxstd.h>
#import <maximize>
#import <solveqp>


//////////// LIML ////////////////////////////
class Liml
{
	decl vy, vx, mZ; decl n;
	decl lambda, betaliml;

	Liml(const y, const x, const Z);
	virtual Estimate();
	fLiml(const vP, const adFunc, const avScore, const amHessian);
	virtual	Coverage();
	GetEstimate();
	GetLambda();
}	

Liml::Liml(const y, const x, const Z)
{
	vy = y; vx = x; mZ = Z;	n=rows(vx);

}

Liml::fLiml(const vP, const adFunc, const avScore, const amHessian)
{
	decl fA, fB, fAd, fBd, fAdd, fBdd;	
	fA = (vy - vP*vx)'mZ*invert(mZ'mZ)*mZ'(vy - vP*vx);
	fB = (vy - vP*vx)'(vy - vP*vx);
	adFunc[0]= - fA / fB;
	
return 1;
}

Liml::Estimate()
{
	decl ir, vp, dfunc;
	vp = 0.5;	// starting value
	ir = MaxBFGS(fLiml, &vp, &dfunc, 0, TRUE);
	betaliml = vp; lambda = - dfunc;
}

Liml::GetEstimate()
{
	return betaliml;
}

Liml::GetLambda()
{
	return lambda;
}

Liml::Coverage()
{
  decl xhat;
  xhat = mZ*invert(mZ'mZ)*mZ'vx - lambda*vx;
  decl vhat;  
  vhat = (1/n)*(vy- vx*betaliml)' (vy- vx*betaliml)
	 * invert(xhat'vx) *(xhat'xhat) *invert(vx'xhat);  
 return (0.1 > ( betaliml - 1.96*sqrt(vhat)) ) 
	    && (0.1 < (betaliml + 1.96*sqrt(vhat)));
}
//////////// Model averagin LIML//////////////////
class MALiml : Liml
{

	decl weight, k, mE, xhat, mK, mGamma;
	decl sigmau, sigmae, sigmaue ; 
	decl mGmin, mGmax, mU;
	decl mxpx, mxpy, mypy, vlam, vbeta;

	MALiml(const y, const x, const Z);
	Firststage();
	Estimate();
	Coverage();
	fMALiml(const vP, const adFunc, const avScore, const amHessian);
 	GetKWp();
	GetKWn();

	SelWDN();
	SelWMA();
	SelWMAC();
	SelWMAU();
}

MALiml::MALiml(const y, const x, const Z)
{
	Liml(y,x,Z); k = columns(mZ); decl limlest;
	mK = range(1,k)';
    mGmin = mK*ones(1,k) .< ones(k,1)*(mK') .? mK*ones(1,k) .: ones(k,1)*(mK');
	mGmax = mK*ones(1,k) .> ones(k,1)*(mK') .? mK*ones(1,k) .: ones(k,1)*(mK');
	mxpx = zeros(k,1); mxpy = zeros(k,1);mypy = zeros(k,1); vlam = zeros(k,1); vbeta = zeros(k,1);
	decl j; for (j=0;j<k;j++)
	{
		mxpx[j] = vx'mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vx;
		mxpy[j] = vx'mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vy;
		mypy[j] = vy'mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vy;
		limlest = new Liml(vy,vx,mZ[][:j]);
		limlest.Estimate();	
		vlam[j] = limlest.GetLambda();
		vbeta[j] = limlest.GetEstimate();
		delete limlest;
	}
	mU = mK*ones(1,k) .> ones(k,1)*(mK') .? mxpx*ones(1,k) .: ones(k,1)*(mxpx');
	mU = vx'vx - mU;
	Firststage();
	
}

MALiml::GetKWp()
{
	return (maxc((weight~zeros(k,1))'))*mK;
}
MALiml::GetKWn()
{
	return fabs((minc((weight~zeros(k,1))'))*mK);
}

MALiml::Firststage()
{
// Mallow's criterion for the first stage //
   sigmau = (vx'vx - vx'mZ*invertsym(mZ'mZ)*mZ'vx)/n;

     decl dnc; dnc=zeros(k,2); dnc[][0] = range(0,k-1)';
     decl j; for (j=0;j<k;j++){
     dnc[j][1]= mU[j][j] + 2*(j+1)*sigmau;
     }
     dnc = sortbyc(dnc,1); decl optk = dnc[0][0];
   decl firstliml; 
   decl delta; delta = vbeta[optk]; 
    decl vxhat; vxhat =  mZ[][:optk]*invertsym(mZ[][:optk]'mZ[][:optk])*(mZ[][:optk]'vx);
 
// minimizing MSE //   
   sigmae = (vy - delta*vx)'(vy - delta*vx)/n; 
   sigmau = (vx - vxhat)'(vx - vxhat)/n;
   sigmaue=(vx - vxhat)'(vy - delta*vx)/n; 

}

MALiml::fMALiml(const vP, const adFunc, const avScore, const amHessian)
{
	decl fA, fB;	
	fA=0;
	fA = weight'(mypy -2*vP*mxpy + vP*vP*mxpx);
	fB = (vy - vP*vx)'(vy - vP*vx);
	adFunc[0]= - fA / fB;
	
return 1;
}


MALiml::Estimate()
{
	lambda = weight'vlam;
	betaliml =(weight'mxpy - lambda*vx'vy)/ (weight'mxpx - lambda*vx'vx);
}


MALiml::Coverage()
{
	xhat =0;
	decl j; for (j=0;j<k;j++){
	xhat = xhat +
	weight[j]*mZ[][:j]*invertsym(mZ[][:j]'mZ[][:j])*mZ[][:j]'vx; 
	}
	xhat = xhat - lambda*vx;
  decl vhat; decl n; n = rows(vx);   
  vhat = (1/n)*(vy- vx*betaliml)' (vy- vx*betaliml)
	 * invert(xhat'vx) *(xhat'xhat) *invert(vx'xhat);  
 return (0.1 > ( betaliml - 1.96*sqrt(vhat)) ) 
	    && (0.1 < (betaliml + 1.96*sqrt(vhat)));

}

////

MALiml::SelWDN()
{
 decl dnc; dnc=zeros(k,2); dnc[][0] = range(0,k-1)';
 decl j;
   for (j=0; j<k ; j++) {
       dnc[j][1] = (2*sigmae*sigmau-sigmaue^2)*(j+1) + sigmae*(mU[j][j] - (mU[k-1][k-1]) );
       }
   dnc = sortbyc(dnc,1); 
   decl optk = dnc[0][0];
   weight = zeros(k,1); weight[optk]=1;
   return vbeta[optk];
}
////


/////////////MA-C /////////////////////
MALiml::SelWMAC()
{


    decl vlo = -ones(k,1); decl vhi = ones(k,1);
    decl ir;
    decl mG, vG, mC, vD, mA, vB;

    mG = 2*((-sigmaue^2)*mGmin + sigmae*mU - sigmae*mU[k-1][k-1]);
    vG = 2*sigmae*sigmau*mK;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
 
  [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,vlo,vhi);

}

//////// MA-P ///////////////////
MALiml::SelWMA()
{


    decl vlo = zeros(k,1); decl vhi = ones(k,1);
    decl ir;
    decl mG, vG, mC, vD;

    mG = 2*((-sigmaue^2)*mGmin + sigmae*mU - sigmae*mU[k-1][k-1]);
    vG = 2*sigmae*sigmau*mK;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
 
   [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,vlo,vhi);

}

//////// MA-U ///////////////////
MALiml::SelWMAU()
{


    decl ir;
    decl mG, vG, mC, vD, mA, vB;

    mG = 2*((-sigmaue^2)*mGmin + sigmae*mU - sigmae*mU[k-1][k-1]);
    vG = 2*sigmae*sigmau*mK;
    mC = ones(1,k);
    vD =1; 
    weight = ones(k,1)/k;	// starting value
	
   [ir,weight]= SolveQP(mG,vG,<>,<>,mC,vD,<>,<>);

}


////////////// simulation class ///////

class Monte
{
	decl n, k, c, Rf2,model;
	decl monte, mSigma, beta, gamma;
	decl betaest, covest, kwest, kwnest;
	
	Monte(const sn, const sk, const sc, const r, const smodel);
	Experiment();
	GetResult();
}

Monte::Monte(const sn, const sk, const sc, const r, const smodel)
{
	n = sn; k=sk; c = sc; Rf2 = r; model=smodel;
	monte =5000;
	mSigma = ones(2,2); mSigma[0][1] = c; mSigma [1][0] =c;
	beta =0.1;

	if 	(model == 1) {
	// part 1 //
	 gamma = ones(k,1)*sqrt(Rf2/(k*(1-Rf2 ) ) ) ;
	}
	else if (model ==2)	{
	//part 2 //
	  gamma=zeros(k,1); decl j; 
	  for (j=0;j<k;j++){ gamma[j]=(1- (j/(k+1)))^4; }
	gamma = sqrt(Rf2/((1-Rf2)*(gamma'gamma)) )*gamma;  
	}
	else if (model == 3){
	// part 3 //
	gamma = zeros(k,1);	decl j; 
	  for (j=0;j<k/2;j++){ gamma[j+k/2]=(1- (j/(k/2+1)))^4; }
	gamma = sqrt(Rf2/((1-Rf2)*(gamma'gamma)) )*gamma; 
	}

}

Monte::Experiment()
{
	ranseed(23587608);

// parameter values //
       decl vy, vx, mZ, mU; 
       decl mSigmac; mSigmac = choleski(mSigma); 

print("gammas", gamma'); 
print ("n=" , n, " K= ", k , " c=",mSigma[1][0] , " Rf2 = ", Rf2  );

// monte carlo //

	decl betaliml = zeros(monte,1);
	decl betadn = zeros(monte,1);
	decl betamau = zeros(monte,1);
	decl betamac = zeros(monte,1);
	decl betamap = zeros(monte,1);

	decl kwdn = zeros(monte,1);
	decl kwmau = zeros(monte,1);
	decl kwmac = zeros(monte,1);
	decl kwmap = zeros(monte,1);

	decl kwmaun = zeros(monte,1);
	decl kwmacn = zeros(monte,1);

   
decl i; for(i=0;i<monte;i++){

// DGP //
       mZ = rann(n,k); mU = rann(n,2)*(mSigmac');
       vx = mZ*gamma + mU[][1];
       vy = beta*vx  + mU[][0]; 
// LIML ESTIMATION //
       decl mybeta; 
       mybeta = new Liml(vy,vx,mZ);
       mybeta.Estimate();
       betaliml[i] = mybeta.GetEstimate();
       delete mybeta;

// DNLIML ESTIMATION //
       mybeta = new MALiml(vy,vx,mZ);
	   mybeta.SelWDN();
	   mybeta.Estimate();
       betadn[i] = mybeta.GetEstimate();
	   kwdn[i] = mybeta.GetKWp();
	   
// MALIML ESTIMATION (U)//
	   mybeta.SelWMAU();	   
       mybeta.Estimate();
       betamau[i] = mybeta.GetEstimate();
	   kwmau[i] = mybeta.GetKWp();
	   kwmaun[i] = mybeta.GetKWn();

// MALIML ESTIMATION (C)//
	   mybeta.SelWMAC();	   
       mybeta.Estimate();
       betamac[i] = mybeta.GetEstimate();
	   kwmac[i] = mybeta.GetKWp();
	   kwmacn[i] = mybeta.GetKWn();

// MALIML ESTIMATION (P)//
	   mybeta.SelWMA();	   
       mybeta.Estimate();
       betamap[i] = mybeta.GetEstimate();
	   kwmap[i] = mybeta.GetKWp();
       delete mybeta;

}

betaest = betaliml~betadn~betamau~betamac~betamap;
kwest = kwdn~kwmau~kwmac~kwmap;
kwnest = kwmaun~kwmacn;

}

Monte::GetResult()
{

	decl mad =  quantilec( fabs((betaest) -beta));
	decl iqr = quantilec(betaest,0.75) - quantilec(betaest,0.25);
	decl maddn = mad/mad[][1];
	decl ressta = (quantilec(betaest)-beta)|iqr|mad|maddn|(k~meanc(kwest))|(zeros(1,2)~meanc(kwnest)~zeros(1,1));

	print( "%r", {"median bias","IQR","MAD","RMAD","KW+","KW-"},
	       "%c",
	{"Liml", "DNLiml","MA-U","MA-C","MA-P"},
		"%7.3g" ,ressta);
}

//////////// MAIN PROGRAM ///////////
main()
{
	decl na,ma,ca,r2a,moda; decl myMonte;
	na=<100>; ma = <20>;ca=<0.1;0.5;0.9>; r2a = <0.1;0.01>; moda=<1;2;3>;
		decl ni,mi,ci,r2i,modi;
		
		for (ni=0; ni < rows(na);ni++){
		for (modi=0; modi < rows(moda);modi++){
		for (r2i=0; r2i < rows(r2a);r2i++){
		for (mi=0; mi < rows(ma);mi++){
		for (ci=0; ci < rows(ca);ci++){
	myMonte = new Monte(na[ni],ma[mi],ca[ci],r2a[r2i],moda[modi]);
	myMonte.Experiment();
	myMonte.GetResult();
	delete myMonte;
	}
	}
	}
	}
    }

	na=<1000>; ma = <30>;ca=<0.1;0.5;0.9>; r2a = <0.1;0.01>; moda=<1;2;3>;
		for (ni=0; ni < rows(na);ni++){
		for (modi=0; modi < rows(moda);modi++){
		for (r2i=0; r2i < rows(r2a);r2i++){
		for (mi=0; mi < rows(ma);mi++){
		for (ci=0; ci < rows(ca);ci++){
	myMonte = new Monte(na[ni],ma[mi],ca[ci],r2a[r2i],moda[modi]);
	myMonte.Experiment();
	myMonte.GetResult();
	delete myMonte;
	}
	}
	}
	}
	}
	
}